"""Core module for defining global settings of the Vantage Agent."""

import re
import sys
from pathlib import Path, PosixPath
from typing import Annotated, Optional

from pydantic import AnyHttpUrl, Field, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from vantage_agent.logger import logger


def _define_dotenv_file_location() -> PosixPath:
    """Define the location of the .env file based on the env var `DOTENV_FILE_LOCATION`.

    In case it is not set, the default location is defined as `.env`.
    """
    default_dotenv_file_location = Path("/var/snap/vantage-agent/common/.env")
    if default_dotenv_file_location.exists():
        return default_dotenv_file_location
    return Path(".env")


class Settings(BaseSettings):
    """Settings for the Vantage Agent."""

    PARTITIONS_JSON_PATH: Path = Path("/nfs/slurm/etc/aws/partitions.json")
    SCONTROL_PATH: Path = Path("/usr/bin/scontrol")
    SLURM_CONF_PATH: Path = Path("/etc/slurm/slurm.conf")

    IS_CLOUD_CLUSTER: bool = False

    # Vantage API info
    BASE_API_URL: Annotated[str, AnyHttpUrl] = "https://apis.vantagehpc.io"

    # Sentry
    SENTRY_DSN: Optional[AnyHttpUrl] = None
    SENTRY_ENV: str = "local"

    # OIDC config for machine-to-machine security
    OIDC_DOMAIN: str
    OIDC_CLIENT_ID: str
    OIDC_CLIENT_SECRET: str
    OIDC_USE_HTTPS: bool = True
    CLUSTER_NAME: str

    CACHE_DIR: Path = Path.home() / ".cache/vantage-agent"

    # Task settings
    TASK_JOBS_INTERVAL_SECONDS: int = Field(30, ge=10, le=3600)  # seconds
    TASK_SELF_UPDATE_INTERVAL_SECONDS: Optional[int] = Field(None, ge=0)  # seconds

    @model_validator(mode="before")
    @classmethod
    def validade_cluster_name(cls, values):
        """Pre validator to set default values."""
        if not values.get("CLUSTER_NAME") and values.get("OIDC_CLIENT_ID"):
            client_regex = (
                r"^(.*)-([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})$"  # noqa
            )
            client_id_regex = re.compile(client_regex)
            match = client_id_regex.match(values.get("OIDC_CLIENT_ID"))
            if match is not None:
                cluster_name = match.group(1)
                values.update(
                    CLUSTER_NAME=cluster_name,
                )
        return values

    @field_validator("SCONTROL_PATH", mode="after")
    @classmethod
    def validate_scontrol_path(cls, v: Path) -> Path:
        """Ensure that the SCONTROL_PATH is an absolute path."""
        if not v.is_absolute():
            raise ValueError("SCONTROL_PATH must be an absolute path")
        return v

    model_config = SettingsConfigDict(
        env_prefix="VANTAGE_AGENT_", env_file=_define_dotenv_file_location(), extra="ignore"
    )


def _init_settings() -> Settings:
    try:
        return Settings()  # type: ignore[call-arg]
    except ValueError as e:
        logger.error(e)
        sys.exit(1)


SETTINGS = _init_settings()



"""Provide to the agent the ability to load custom plugins that are installed on the same environment."""

from importlib.metadata import entry_points
from typing import Any, Dict

from buzz import handle_errors

from vantage_agent.logger import logger, logger_wraps


@logger_wraps()
def load_plugins(plugin_name: str) -> Dict[str, Any]:
    """Discover and load plugins available to the agent, allowing for third party ones to be included.

    Notice the ones shipped with the agent are also declared on the ``pyproject.toml`` file
    as plugins, even though they could be easily loaded directly from source. This aims
    to support tests and to demonstrate how to use the plugin system.

    Reference:
        https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/
    """
    discovered_plugins = entry_points(group=f"vantage_agent.{plugin_name}")
    logger.info(f"Loaded plugins {discovered_plugins}")
    result = {}
    for plugin in discovered_plugins:
        with handle_errors(
            f"Failed to load plugin {plugin_name} from {plugin.value}",
            raise_exc_class=RuntimeError,
        ):
            content = plugin.load()

        result[plugin.name.lower()] = content

    logger.info("Discovered the following plugins for {}: {}", plugin_name, ", ".join(result.keys()))
    return result



"""Core module for defining operations related to calling the Vantage API."""

from textwrap import dedent
from typing import Any, TypedDict

from loguru import logger

from vantage_agent.settings import SETTINGS
from vantage_agent.vantage_api_client import AsyncBackendClient


class UploadSlurmConfigSuccess(TypedDict):
    """Type definition for the success response of the uploadSlurmConfig mutation."""

    data: dict[Any, Any]
    errors: list[Any] | None


async def upload_slurm_config(slurm_config: dict[Any, Any]) -> None:
    """Upload Slurm's configuration to the Vantage API."""
    query = dedent(
        """
        mutation($config: JSONScalar!, $clientId: String!) {
            uploadSlurmConfig(config: $config, clientId: $clientId) {
                ... on UploadSlurmConfigSuccess {
                    message
                }
            }
        }
        """
    )
    body = {"query": query, "variables": {"config": slurm_config, "clientId": SETTINGS.OIDC_CLIENT_ID}}
    async with AsyncBackendClient() as backend_client:
        res = await backend_client.post("/cluster/graphql", json=body)

    response_data: UploadSlurmConfigSuccess = res.json()
    logger.debug(f"Response from Vantage API while uploading Slurm's configurations: {response_data}")


async def upload_slurm_partitions(slurm_partitions: dict[Any, Any]) -> None:
    """Upload Slurm's partitions information to the Vantage API."""
    query = dedent(
        """
        mutation($partitions: JSONScalar!, $clientId: String!) {
            uploadSlurmPartitions(partitions: $partitions, clientId: $clientId) {
                ... on UploadSlurmPartitionsSuccess {
                    message
                }
            }
        }
        """
    )
    body = {
        "query": query,
        "variables": {"partitions": slurm_partitions, "clientId": SETTINGS.OIDC_CLIENT_ID},
    }
    async with AsyncBackendClient() as backend_client:
        res = await backend_client.post("/cluster/graphql", json=body)

    response_data: UploadSlurmConfigSuccess = res.json()
    logger.debug(f"Response from Vantage API while uploading partitions: {response_data}")


async def upload_slurm_nodes(slurm_nodes: dict[Any, Any]) -> None:
    """Upload Slurm's nodes information to the Vantage API."""
    query = dedent(
        """
        mutation($nodes: JSONScalar!, $clientId: String!) {
            uploadSlurmNodes(nodes: $nodes, clientId: $clientId) {
                ... on UploadSlurmNodesSuccess {
                    message
                }
            }
        }
        """
    )
    body = {"query": query, "variables": {"nodes": slurm_nodes, "clientId": SETTINGS.OIDC_CLIENT_ID}}
    async with AsyncBackendClient() as backend_client:
        res = await backend_client.post("/cluster/graphql", json=body)

    response_data: UploadSlurmConfigSuccess = res.json()
    logger.debug(f"Response from Vantage API while uploading nodes: {response_data}")


async def clean_slurm_partitions(slurm_partitions: list[str] | set[str]) -> None:
    """Clean up partitions that are no longer present in Slurm."""
    query = dedent(
        """
        mutation($partitions: [String!]!, $clientId: String!) {
            cleanSlurmPartitions(partitions: $partitions, clientId: $clientId) {
                ... on CleanSlurmPartitionsSuccess {
                    message
                }
            }
        }
        """
    )
    body = {
        "query": query,
        "variables": {"partitions": slurm_partitions, "clientId": SETTINGS.OIDC_CLIENT_ID},
    }
    async with AsyncBackendClient() as backend_client:
        res = await backend_client.post("/cluster/graphql", json=body)

    response_data: UploadSlurmConfigSuccess = res.json()
    logger.debug(f"Response from Vantage API while cleaning partitions: {response_data}")


async def clean_slurm_nodes(slurm_nodes: list[str] | set[str]) -> None:
    """Clean up nodes that are no longer present in Slurm."""
    query = dedent(
        """
        mutation($nodes: [String!]!, $clientId: String!) {
            cleanSlurmNodes(nodes: $nodes, clientId: $clientId) {
                ... on CleanSlurmNodesSuccess {
                    message
                }
            }
        }
        """
    )
    body = {"query": query, "variables": {"nodes": slurm_nodes, "clientId": SETTINGS.OIDC_CLIENT_ID}}
    async with AsyncBackendClient() as backend_client:
        res = await backend_client.post("/cluster/graphql", json=body)

    response_data: UploadSlurmConfigSuccess = res.json()
    logger.debug(f"Response from Vantage API while cleaning nodes: {response_data}")



"""Core module for defining partitions related operations."""

import json
import subprocess
from shutil import copyfile
from textwrap import dedent

from vantage_agent.logger import logger
from vantage_agent.settings import SETTINGS
from vantage_agent.vantage_api_client import AsyncBackendClient


def create_backup(file_path: str):
    """Create a backup of slurm.conf."""
    bkp_name = str(file_path).split("/").pop()
    backup_path = f"{SETTINGS.CACHE_DIR}/{bkp_name}.bkp"
    copyfile(file_path, backup_path)
    logger.info(f"Backup of {bkp_name} created at {backup_path}")


async def update_slurm_config(partitions):
    """Update the slurm.conf file with the given partitions."""
    # Backup slurm.conf file
    create_backup(SETTINGS.SLURM_CONF_PATH)
    logger.debug(f"Current SLURM path: {SETTINGS.SLURM_CONF_PATH}")

    with open(SETTINGS.SLURM_CONF_PATH, "r") as f:
        slurm_config = f.readlines()

    logger.debug(f"Current SLURM config: {slurm_config}")
    new_partition_configs = []
    node_names_seen = set()

    for partition in partitions:
        logger.debug(f"Processing partition: {partition}. Type: {type(partition)}")
        instance_info = await get_node_type_info(partition["nodeType"])

        # Build the new config for each partition
        is_default = "Yes" if partition.get("isDefault", None) else "No"
        node_range = f"[0-{partition['maxNodeCount'] - 1}]"
        node_name = f"NodeName={partition['name']}-{partition['name']}-{node_range} State=CLOUD Weight=1 Feature=cloud CPUs={instance_info['numCpus']}\n"  # noqa
        partition_name = f"PartitionName={partition['name']} Nodes={partition['name']}-{partition['name']}-{node_range} MaxNodes={partition['maxNodeCount']} MaxTime=INFINITE State=UP Default={is_default}\n"  # noqa

        new_partition_configs.append(node_name)
        new_partition_configs.append(partition_name)

        # Toggle node as done
        node_names_seen.add(partition["name"])

    # Filter the lines that aren't Partition or Node configs
    filtered_config = []
    for line in slurm_config:
        if line.startswith("NodeName=") or line.startswith("PartitionName="):
            continue
        filtered_config.append(line)

    logger.debug(f"Filtered lines from SLURM {filtered_config}")

    # Add the new partitions to the end of slurm.conf file and write a new file
    filtered_config += new_partition_configs
    with open(SETTINGS.SLURM_CONF_PATH, "w") as f:
        f.writelines(filtered_config)
    # Copy file to the nfs directory
    copyfile(SETTINGS.SLURM_CONF_PATH, "/nfs/slurm/etc/slurm/slurm.conf")

    node_names = ", ".join(node_names_seen)
    logger.info(f"SLURM conf file updated with the partitions {node_names}.")


async def update_partition_config(partitions):
    """Update the partitions.json file with the given partitions."""
    create_backup(SETTINGS.PARTITIONS_JSON_PATH)
    logger.debug(f"Current partitions.json path: {SETTINGS.PARTITIONS_JSON_PATH}")

    with open(SETTINGS.PARTITIONS_JSON_PATH, "r", encoding="utf-8") as f:
        partitions_config = json.load(f)

    new_partitions = []
    for partition in partitions:
        is_default = "Yes" if partition.get("isDefault", None) else "No"
        instance_info = await get_node_type_info(partition["nodeType"])
        new_partition = {
            "PartitionName": partition["name"],
            "NodeGroups": [
                {
                    "NodeGroupName": partition["name"],
                    "MaxNodes": partition["maxNodeCount"],
                    "Region": partitions_config["Partitions"][0]["NodeGroups"][0]["Region"],
                    "SlurmSpecifications": {
                        "Weight": 1,
                        "Feature": "cloud",
                        "CPUs": instance_info["numCpus"],
                    },
                    "PurchasingOption": "on-demand",
                    "OnDemandOptions": {"AllocationStrategy": "lowest-price"},
                    "LaunchTemplateSpecification": {
                        "LaunchTemplateId": partitions_config["Partitions"][0]["NodeGroups"][0][
                            "LaunchTemplateSpecification"
                        ]["LaunchTemplateId"],  # noqa
                        "Version": "$Latest",
                    },
                    "LaunchTemplateOverrides": [{"InstanceType": partition["nodeType"]}],
                    "SubnetIds": partitions_config["Partitions"][0]["NodeGroups"][0]["SubnetIds"],
                }
            ],
            "PartitionOptions": {"Default": is_default},
        }
        new_partitions.append(new_partition)

    with open(SETTINGS.PARTITIONS_JSON_PATH, "w", encoding="utf-8") as f:
        json.dump({"Partitions": new_partitions}, f, indent=4, ensure_ascii=False)


async def get_node_type_info(instance_type: str):
    """Get the node type specifications from the api."""
    query = dedent(
        """
        query($filters: JSONScalar!, $first: Int, $after: Int) {
            awsNodePicker(filters: $filters, after: $after, first: $first) {
                edges {
                    node {
                        awsRegion
                        cpuName
                        cpuManufacturer
                        cpuArch
                        gpuName
                        gpuManufacturer
                        id
                        instanceType
                        memory
                        numCpus
                        numGpus
                        pricePerHour
                    }
                }
            }
        }
        """
    )

    body = {
        "query": query,
        "variables": {"filters": {"instanceType": {"eq": instance_type}}, "after": 1, "first": 100},
    }

    async with AsyncBackendClient() as backend_client:
        res = await backend_client.post("/cluster/graphql", json=body)
        res = res.json()
    instance_info = res["data"]["awsNodePicker"]["edges"][0]["node"]
    return instance_info


async def get_cluster_partitions():
    """Get the partitions from the api."""
    query = dedent(
        """
        query($filters: JSONScalar!, $first: Int, $after: Int) {
            partitions(filters: $filters, after: $after, first: $first) {
                edges {
                    node {
                        clusterName
                        id
                        isDefault
                        maxNodeCount
                        name
                        nodeType
                    }
                }
            }
        }
        """
    )

    body = {
        "query": query,
        "variables": {"filters": {"clusterName": {"eq": SETTINGS.CLUSTER_NAME}}, "after": 1, "first": 100},
    }

    async with AsyncBackendClient() as backend_client:
        res = await backend_client.post("/cluster/graphql", json=body)
        res = res.json()
    partitions = res["data"]["partitions"]["edges"]
    return [partition["node"] for partition in partitions]


def reconfigure_slurm():
    """Reload the slurm.conf to get the new changes in the slurm.conf."""
    try:
        logger.debug("Reloading SLURM config.")
        result = subprocess.run(
            ["systemctl", "restart", "slurmctld"], check=True, capture_output=True, text=True
        )
        logger.info(f"SLURM config reloaded:\n{result.stdout}")
    except subprocess.CalledProcessError as e:
        logger.error(f"Error while reloading SLURM conf:\n{e.stderr}")
        raise e


async def sync_cluster_partitions():
    """Sync the slurm partitions with the current partitions in the API."""
    try:
        logger.info("Syncing SLURM partitions.")

        partitions = await get_cluster_partitions()

        logger.debug(f"Updating SLURM config file with partition: {partitions}.")
        await update_slurm_config(partitions=partitions)

        logger.debug(f"Updating partitions.json config file with partition: {partitions}.")
        await update_partition_config(partitions=partitions)

        logger.debug("Reloading SLURM config.")
        reconfigure_slurm()

        logger.info("SLURM partitions synced.")
    except Exception as e:
        logger.error(f"Error while syncing the partitions: {str(e)}")



"""Core module for Sentry related operations.

Copied from https://github.com/omnivector-solutions/jobbergate/blob/main/jobbergate-agent/jobbergate_agent/utils/sentry.py.
"""

import logging

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.utils import BadDsn

from vantage_agent.logger import logger
from vantage_agent.settings import SETTINGS


def init_sentry():
    """Initialize the Sentry SDK."""
    try:
        sentry_logging = LoggingIntegration(level=logging.WARNING, event_level=logging.ERROR)

        sentry_sdk.init(
            dsn=SETTINGS.SENTRY_DSN,
            integrations=[sentry_logging],
            traces_sample_rate=1.0,
            environment=SETTINGS.SENTRY_ENV,
        )

        logger.debug("Enabled Sentry since a valid DSN key was provided.")
    except BadDsn as e:
        logger.debug(f"Sentry could not be enabled: {e}")
